## Copyright 202f VMware, Inc.  All rights reserved. VMware Confidential.
## This cleanup script is idempotent.  It is run during first deployment and subsequent updates.
import json
import logging
import os
import shlex
import subprocess
import sys
from kubernetes import client, config, stream
from subprocess import CalledProcessError, STDOUT
from time import sleep

# Set up environment variables for POSTGRESSQL calls
my_env = os.environ.copy()

# TODO we may get password form a file, at least based on the helm chart.
my_env['PGPASSWORD'] = my_env.get('POSTGRES_PASSWORD', 'postgres')

# Define global variables
NAMESPACE = my_env.get('NAMESPACE')
RES_NAME = my_env.get('RESOURCE_NAME')
PSQL_COMPONENT = 'postgresql'
PGPOOL_COMPONENT = 'pgpool'
PSQL_USERNAME = my_env.get('POSTGRES_USERNAME')
INSTANCE = my_env.get('INSTANCE_NAME')
PSQL_STS_NAME = my_env.get('PSQL_STATEFULSET_NAME')
PGPOOL_DEPLOYMENT_NAME = my_env.get('PGPOOL_DEPLOYMENT_NAME')
SERVICE_ENDPOINT_ENV_NAME = my_env.get("PSQL_SERVICE_ENV_NAME")
SERVICE_ENDPOINT_HOST = my_env.get(SERVICE_ENDPOINT_ENV_NAME)
LABEL_SELECTOR = 'app.kubernetes.io/component=%s,app.kubernetes.io/name=%s,app.kubernetes.io/instance=%s' % (PSQL_COMPONENT, RES_NAME, INSTANCE)
REPMGR_SLOT_PREFIX = 'repmgr_slot_1'
PSQL0_POD_NAME = PSQL_STS_NAME + '-0'
LOG_FORMAT = '%(asctime)-15s %(levelname)s %(message)s'

# Create logger
logging.basicConfig(format=LOG_FORMAT)
logger = logging.getLogger("PostgresCleanup")
logger.setLevel(logging.INFO)

# Setup K8S API clients
config.load_incluster_config()
core_v1_api = client.CoreV1Api()
apps_v1_api = client.AppsV1Api()

def subprocess_output_w_returncode(cmd_str, timeout=30):
    try:
        output = subprocess.check_output(shlex.split(cmd_str), env=my_env, timeout=timeout, stderr=STDOUT, encoding='utf-8')
        return output.strip(), 0
    except subprocess.CalledProcessError as grepexc:
        return grepexc.output.strip(), grepexc.returncode


def parse_psql_output(s: str):
    """
    Return Json object embedded in 's'
    :param s: input strings with json string in it
    :return: json object
    """
    b = s.find('[')
    e = s.rfind(']')
    if b != -1 and e != -1 and b < e:
        json_str = s[b:e+1]
        jstr_lines = json_str.split('\n')
        jstr_lines = [l.strip('+') for l in jstr_lines]
        return json.loads(''.join(jstr_lines))
    return None


def get_repmgr_status():
    repmgr_cmd = ['repmgr', '-f', '/opt/build/repmgr/conf/repmgr.conf',
                  'service', 'status']
    # Kubectl exec on postgres node 0
    out = stream.stream(core_v1_api.connect_get_namespaced_pod_exec,  PSQL0_POD_NAME,
                        NAMESPACE, command=repmgr_cmd, container='postgresql',
                        stderr=True, stdout=True, stdin=False, tty=False)
    logger.info("Repmgr service status:\n %s", out)


def get_replication_slots():
    cmdline = 'psql -w -q -U "%s" -d "postgres" -h %s -c "SELECT json_agg(pg_replication_slots) FROM pg_replication_slots;"' % (PSQL_USERNAME, SERVICE_ENDPOINT_HOST)
    output, c = subprocess_output_w_returncode(cmdline)
    if c != 0:
        logger.error("Failed to get replication slots from %s - %s, %s", SERVICE_ENDPOINT_HOST, slot_name, c, output)
        sys.exit(1)
    logger.info("Available replication slots: %s", output)
    return parse_psql_output(output)


def drop_replication_slot(slot_name):
    logger.debug("Dropping replication slot '%s'", slot_name)

    psql_cmdline = 'psql -w -q -U "%s" -d "postgres" -h %s -c "SELECT pg_drop_replication_slot(\'%s\');"' % (PSQL_USERNAME, SERVICE_ENDPOINT_HOST, slot_name)
    output, c = subprocess_output_w_returncode(psql_cmdline)
    if c != 0:
        logger.error("Failed to drop pg replication slot %s from %s - %s, %s", slot_name, SERVICE_ENDPOINT_HOST, c, output)
        sys.exit(1)
    else:
        logger.info("Pg replication slot '%s' dropped, result: %s", slot_name, output)

    repmgr_slot_id = slot_name[slot_name.rfind('_') + 1:]
    repmgr_cmd = ['repmgr', '-f', '/opt/build/repmgr/conf/repmgr.conf',
                  'standby', 'unregister', '--node-id', repmgr_slot_id ]
    # Kubectl exec on postgres node 0
    out = stream.stream(core_v1_api.connect_get_namespaced_pod_exec,  PSQL0_POD_NAME,
                         NAMESPACE, command=repmgr_cmd, container='postgresql',
                         stderr=True, stdout=True, stdin=False, tty=False)
    logger.info("Unregistering repmgr node %s output:\n %s",
                repmgr_slot_id, out)


def get_pgpool_deployment_update_status():
    """
    Return True if the target deployment has completed upgrading of all pods and
    the number of ready pods are consistent to the desired number
    """
    ds = apps_v1_api.read_namespaced_deployment(PGPOOL_DEPLOYMENT_NAME, NAMESPACE)
    logger.debug("Got deployment for '%s'", ds.metadata.name)
    logger.info("Current status of PGPool deployment %s: spec replicas: %s, updated replicas:"
                " %s, ready replicas: %s", PGPOOL_DEPLOYMENT_NAME, ds.spec.replicas,
                ds.status.updated_replicas, ds.status.ready_replicas)
    return ds.status.updated_replicas == ds.spec.replicas \
           and ds.status.updated_replicas == ds.status.ready_replicas

def get_psql_stateful_set_update_status():
    """
    Return True if the target stateful set has completed upgrading of all pods and
    the number of ready pods are consistent to the desired number
    """
    global psql_sts
    psql_sts  = apps_v1_api.read_namespaced_stateful_set(PSQL_STS_NAME, NAMESPACE)
    logger.debug("Got target stateful set for '%s'", psql_sts.metadata.name)
    logger.info("Current status of Postgresql stateful set %s: spec replicas: %s, updated replicas:"
                " %s, ready replicas: %s", PSQL_STS_NAME, psql_sts.spec.replicas,
                psql_sts.status.updated_replicas, psql_sts.status.ready_replicas)
    return psql_sts.status.updated_replicas == psql_sts.spec.replicas \
           and psql_sts.status.updated_replicas == psql_sts.status.ready_replicas


def delete_pvc_and_pvs(pvcs_names):
    logger.debug("Going to delete PVCs and PVs: %s",
                 pvcs_names)
    for name in pvcs_names:
        pvc = core_v1_api.read_namespaced_persistent_volume_claim(name, NAMESPACE)
        pv_name = pvc.spec.volume_name
        logger.info("PVC '%s' and its backing PV '%s' will be deleted",
                    name, pv_name)
        core_v1_api.delete_namespaced_persistent_volume_claim(name, NAMESPACE)
        core_v1_api.delete_persistent_volume(pv_name)



# Wait for pgpool deployment in question to finish deploying/updating...
while not get_pgpool_deployment_update_status():
    logger.info(
        "PGPool deployment '%s' is still deploying/upgrading!  Wait for 60 secs and re-check",
        PGPOOL_DEPLOYMENT_NAME)
    sleep(60)

# Wait for postgresql stateful set in question to finish deploying/updating...
while not get_psql_stateful_set_update_status():
    logger.info(
        "Postgresql stateful set '%s' is still deploying/upgrading!  Wait for 60 secs and re-check",
        PSQL_STS_NAME)
    sleep(60)

logger.info("Postgresql statefule set %s has finish upgrading.  Perform replication "
            "slots and persistent volume clean up after 3 mins.", PSQL_STS_NAME)
sleep(180)
new_replica_counts = psql_sts.spec.replicas

pg_replication_slots = get_replication_slots()
slots_to_drop = set()
all_slot_names = []

# check for any inactive physical replication slots used by REPMGR_SLOT and
# not part of the cluster
if pg_replication_slots:
    for slot in pg_replication_slots:
        slot_name = slot.get('slot_name')
        all_slot_names.append(slot_name)
        if not slot.get('active') and slot.get('slot_type') == 'physical':
            if slot_name.startswith(REPMGR_SLOT_PREFIX) \
                    and int(slot_name[len(REPMGR_SLOT_PREFIX):]) >= new_replica_counts:
                slots_to_drop.add(slot_name)

if len(all_slot_names) < new_replica_counts - 1:
    logger.error("Expecting %s replication slots but get %s - Failing the job and retry.",
                 new_replica_counts -1, all_slot_names)
    sys.exit(1)

logger.info("All rep slots: %s; slots_to_drop: %s", all_slot_names, slots_to_drop)
# remove unused slots
for slot in slots_to_drop:
    drop_replication_slot(slot)

# Now remove unused PVC's and associate PV's
# Get a list of PVC's labeled for target stateful set
all_pvcs = set()
pvcs = core_v1_api.list_namespaced_persistent_volume_claim(
    NAMESPACE, label_selector=LABEL_SELECTOR)
all_pvcs.update([pvc.metadata.name for pvc in pvcs.items])
if not all_pvcs:
    logger.error("Check system status! No PVC is found using selector '%s', which is not possible", LABEL_SELECTOR)
    sys.exit(1)
logger.info("All PVCs found: %s", all_pvcs)

# Filter out those PVCs in use
pods = core_v1_api.list_namespaced_pod(NAMESPACE, label_selector=LABEL_SELECTOR)
pvcs_in_use = set()
v_inuse = [v.persistent_volume_claim.claim_name
           for p in pods.items
           for v in p.spec.volumes if v.persistent_volume_claim]
pvcs_in_use.update(v_inuse)
logger.info("PVCs in use: %s", pvcs_in_use)
pvcs_to_delete = all_pvcs.difference(pvcs_in_use)

if pvcs_to_delete:
    delete_pvc_and_pvs(pvcs_to_delete)

get_repmgr_status()

logger.info("Repmgr/PVC cleanup completed successfully.")
