/* Function returns the current disk usage percentage of the Postgres data dir */
CREATE OR REPLACE FUNCTION get_disk_usage_percentage() RETURNS integer as $$
DECLARE
    disk_usage integer;
BEGIN
CREATE TEMP TABLE IF NOT EXISTS tmp_sys_df (content text) ON COMMIT DROP;
/* Get the df output from the OS and extract the percentage usage value from the same */
COPY tmp_sys_df FROM PROGRAM 'OUTPUT=$(df $PGDATA | tail -n +2) && echo $OUTPUT';
disk_usage=(SELECT SPLIT_PART(SPLIT_PART(content, ' ', 5),'%',1)::integer FROM tmp_sys_df);
RETURN disk_usage;
END;
$$ LANGUAGE plpgsql;

/* Function checks for inactive replication slots and drops them */
CREATE OR REPLACE FUNCTION clear_inactive_replication_slots() RETURNS void as $$
DECLARE
    slot_names varchar;
    node_count integer;
    replication_slots_count integer;
    inactive_replication_slots_count integer;
BEGIN

/* Get the number of nodes in the cluster from the repmgr db*/
CREATE TEMP TABLE IF NOT EXISTS tmp_cluster_nodes (content text) ON COMMIT DROP;
COPY tmp_cluster_nodes FROM PROGRAM 'PGPASSWORD=$REPMGR_PASSWORD psql -w -U $REPMGR_USERNAME -d $REPMGR_DATABASE -h 127.0.0.1 -t -c ''select count(node_id) from nodes;''';
node_count=(SELECT TRIM(content)::integer FROM tmp_cluster_nodes LIMIT 1);
RAISE INFO 'Number of nodes in the cluster %', node_count;

/* Get the total replication slots */
replication_slots_count=(SELECT count(slot_name) from pg_replication_slots);
RAISE INFO 'Number of replication slots %', replication_slots_count;

/* Get the inactive replication slots count */
inactive_replication_slots_count=(SELECT count(slot_name) FROM pg_replication_slots WHERE active_pid IS NULL OR active_pid NOT IN(SELECT pid FROM pg_stat_replication) AND active IS NOT TRUE);
RAISE INFO 'Number of inactive replication slots %', inactive_replication_slots_count;

/* Drop inactive replication slots only if the number of replication slots are more than or equal to the number of nodes in the cluster */
IF replication_slots_count >= node_count THEN
    FOR slot_names IN SELECT slot_name FROM pg_replication_slots WHERE active_pid IS NULL OR active_pid NOT IN(SELECT pid FROM pg_stat_replication) AND active IS NOT TRUE
        LOOP
            RAISE INFO 'Dropping inactive replication slot %', slot_names;
            PERFORM pg_drop_replication_slot(slot_names);
        END LOOP;
ELSE
    IF inactive_replication_slots_count > 0 THEN
        RAISE INFO 'Number of replication slots less than number of nodes. Not dropping the inactive replication slots';
    ELSE
        RAISE INFO 'No inactive replication slots';
    END IF;
END IF;
END;
$$ LANGUAGE plpgsql;

/*Execute the inactive replication slot cleanup upfront once*/
SELECT clear_inactive_replication_slots();